"""Exceptions."""


class ImageSearchException(Exception):
    """Parent exception"""


class ThresholdException(ImageSearchException):
    """Indicates the threshold is invalid."""


class UnknownAlgorithmException(ImageSearchException):
    """Indicates the specified algorithm does not name a known algorithm."""


class UnknownFormatException(ImageSearchException):
    """Indicates the specified format is unknown."""


class HashingException(ImageSearchException):
    """Indicates the path could not be hashed for some reason."""


class CLIException(ImageSearchException):
    """Indicates a problem parsing command line arguments."""


class DoesNotExistException(HashingException):
    """Indicates that a path does not exist."""


class NotReadableException(HashingException):
    """Indicates the path is not a file (for source images) or is not readable (permissions)."""


class UnsupportedImageException(HashingException):
    """Indicates that the file is not supported for decoding."""



"""The main entry point. Invoke as `imagesearch' or `python -m imagesearch'."""
import sys
from typing import List

from .status import ExitStatus
from .exceptions import ImageSearchException
from .cli import PARSER

def main(args: List[str] = sys.argv) -> None:  # pylint: disable=dangerous-default-value
    """Entry point function."""
    try:
        parsed_args = PARSER.parse_args(args[1:])
        parsed_args.command(parsed_args)
    except ImageSearchException as exc:
        sys.stderr.write(f"imagesearch: error: {exc}\n")
        exit_status = ExitStatus.ERROR
    except KeyboardInterrupt:
        sys.stderr.write("imagesearch: error: interrupted\n")
        exit_status = ExitStatus.ERROR_CTRL_C
    else:
        exit_status = ExitStatus.SUCCESS

    sys.exit(exit_status.value)


if __name__ == '__main__':
    main()



"""Functionality around finding duplicate images."""
from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterable, Set, Generator

from imagehash import ImageHash
import attr

from .fingerprint import Algorithm, ImageFingerprint

@attr.s(frozen=True, order=True, auto_attribs=True, kw_only=True)
class Dupe:
    """A set of image paths that hash to the same value with a given algorithm."""
    image_hash: ImageHash
    algorithm: Algorithm
    paths: Set[Path]

    @classmethod
    def find(
            cls,
            search_paths: Iterable[Path],
            algorithm: Algorithm
    ) -> Generator[Dupe, None, None]:
        """
        Yields Dupe objects for image files in the given search_paths.
        """
        hashes: Dict[ImageHash, Set[Path]] = defaultdict(set)

        for image_fingerprint in ImageFingerprint.recurse_paths(search_paths, algorithm):
            hashes[image_fingerprint.image_hash].add(image_fingerprint.path)

        yield from [
            cls(
                image_hash=image_hash,
                algorithm=algorithm,
                paths=paths
            ) for image_hash, paths in hashes.items() if len(paths) > 1
        ]



"""Definition of the argument parser."""
from pathlib import Path
import argparse
from typing import NoReturn

from .. import __version__
from ..fingerprint import Algorithm
from ..exceptions import CLIException
from .command import DupeCommand, CompareCommand, Format

def _add_algo_arg_to_parser(parser: argparse.ArgumentParser) -> None:
    """
    All subcommands of imagesearch take an optional algorithm option. This helper method allows us
    to avoid repeating ourselves.
    """
    parser.add_argument(
        "-a",
        "--algorithm",
        action="store",
        default=Algorithm.AHASH,
        type=Algorithm.from_name,
        help=f"""
        The algorithm to use to fingerprint the images. See the imagehash library for documentation
        on how each works. {Algorithm.all_descriptions()}.
        """
    )


def _add_format_arg_to_parser(parser: argparse.ArgumentParser) -> None:
    """
    All subcommands of imagesearch take an optional format option. This helper method allows us
    to avoid repeating ourselves.
    """
    parser.add_argument(
        "-f",
        "--format",
        action="store",
        default=Format.JSON,
        type=Format.from_name,
        help=f"""
        The format to output results in. Choose from among: {Format.supported_names()}.
        """
    )

class ImageSearchArgumentParser(argparse.ArgumentParser):
    """Arg parser with customized error handling."""
    def error(self, message: str) -> NoReturn:
        """Raises a CLIException."""
        raise CLIException(message)


PARSER = ImageSearchArgumentParser(
    prog="imagesearch",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="""
    Performs various visual matching operations on images.
    """,
)

PARSER.add_argument(
    "--version",
    action="version",
    version=__version__
)

_SUBCOMMAND_PARSER = PARSER.add_subparsers(
    dest="subcommand",
    required=True,
)

_COMPARE_PARSER = _SUBCOMMAND_PARSER.add_parser(
    "compare",
    help="""
    Compares a reference image to other images and returns a measure of visual similarity.
    """
)

_COMPARE_PARSER.add_argument(
    "ref_path",
    type=Path,
    metavar="REF_PATH",
    default=None,
    help="""
    The reference image to compare among the search paths.
    """
)

_COMPARE_PARSER.add_argument(
    "search_paths",
    nargs="+",
    metavar="SEARCH_PATH",
    type=Path,
    help="""
    The path(s) of other images to compare against the reference image. This argument may be
    supplied multiples times. The path may be either a file (for explicit comparison) or a directory
    that will be recursively searched for image files.
    """
)

_add_algo_arg_to_parser(_COMPARE_PARSER)
_add_format_arg_to_parser(_COMPARE_PARSER)

_COMPARE_PARSER.set_defaults(command=CompareCommand.run)

_COMPARE_PARSER.add_argument(
    "-1",
    "--stop-on-first-match",
    action="store_true",
    help="""
    Stop searching immediately after the first image with diff <= threshold is found
    (instead of continuing to search through all search paths), which also means that
    this argument must be used with the threshold argument as well . If multiple
    matches do exist, the result is arbitrary. This argument is intended to search
    through a large set of images where only a single image is desired.
    """
)

_COMPARE_PARSER.add_argument(
    "-t",
    "--threshold",
    type=int,
    default=None,
    help="""
    The the maximum difference at which to consider 2 images as a match. Lower numbers mean closer
    match. This value is subjective and depends on the hash algorithm used, so testing without a
    threshold may be useful to feel for the values. Omit to use an infinite threshold.
    """
)

_DUPE_PARSER = _SUBCOMMAND_PARSER.add_parser(
    "dupe",
    help="""
    Finds images which hash to the same value within the given paths.
    """
)

_DUPE_PARSER.add_argument(
    "search_paths",
    nargs="+",
    metavar="SEARCH_PATH",
    type=Path,
    help="""
    The path(s) to search for visually similar images. This argument may be supplied
    multiples times. The path may be either a file (for explicit comparison) or a
    directory that will be recursively searched for image files.
    """
)

_add_algo_arg_to_parser(_DUPE_PARSER)
_add_format_arg_to_parser(_DUPE_PARSER)

_DUPE_PARSER.set_defaults(command=DupeCommand.run)



"""
Returns the similiarity between a reference image and a set of other images.
"""
__version__ = '0.2.1'

from .fingerprint import Algorithm, ImageFingerprint
from . import exceptions
from .compare import ImageDiff
from .dupe import Dupe



"""Enumeration of formats."""
from __future__ import annotations

from enum import Enum, auto
from typing import Dict

from ..exceptions import UnknownFormatException

class Format(Enum):
    """Formats for output."""
    TEXT = auto()
    JSON = auto()

    @classmethod
    def from_name(cls, name: str) -> Format:
        """Returns an format by name."""
        name_map: Dict[str, Format] = {
            format_.name.lower(): format_ for format_ in list(Format)
        }
        name_normalized = name.lower()
        if name_normalized not in name_map:
            raise UnknownFormatException(f"No format with name {name}")
        return name_map[name_normalized]

    @classmethod
    def supported_names(cls) -> str:
        """Returns a comma-separated string of all the format names."""
        return ", ".join(format_.name.lower() for format_ in list(Format))



"""CLI commands."""
from __future__ import annotations

import sys
from abc import abstractmethod, ABCMeta
import json
from argparse import Namespace
from typing import Generator, List, Generic, TypeVar, Any, Callable

from imagesearch import Dupe, ImageDiff
from imagesearch.exceptions import UnknownFormatException
from imagesearch.cli.format import Format


ItemT = TypeVar("ItemT")


class _FakeGenericABCMeta(ABCMeta):
    """Allows for generic subtyping"""
    def __getitem__(cls, item: Any) -> ABCMeta:
        """
        Generic subscript hack, like Foo[Bar].

        TODO Is this necessary? Like, do I need to do this to generically subclass?
        """
        return cls


class Command(Generic[ItemT], metaclass=_FakeGenericABCMeta):
    """
    Abstract base class for running commands. Sets up subclass requirements for each output format.
    """
    # If a new format is to be supported, add it to the Format enum, create a new abstract class
    # method, and then add that method to the map in output_function_by_format.

    @classmethod
    def output_function_by_format(
            cls,
            format_: Format
    ) -> Callable[[Namespace, Generator[ItemT, None, None]], str]:
        """
        Selects an output function based on format. Raises an UnknownFormatException if format is
        not known.
        """
        formats = {
            Format.JSON: cls.json_output,
            Format.TEXT: cls.text_output,
        }
        if format_ in formats:
            return formats[format_]

        raise UnknownFormatException(f"No format with name {format_}.")

    @classmethod
    @abstractmethod
    def _generate(cls, args: Namespace) -> Generator[ItemT, None, None]:
        """Generate items for consumption by the output methods."""
        raise NotImplementedError("Cannot call this method on base class.")

    @classmethod
    @abstractmethod
    def json_output(cls, args: Namespace, items: Generator[ItemT, None, None]) -> str:
        """Produce a json string for output of the items"""
        raise NotImplementedError("Cannot call this method on base class.")

    @classmethod
    @abstractmethod
    def text_output(cls, args: Namespace, items: Generator[ItemT, None, None]) -> str:
        """Produce a text string for output of the items"""
        raise NotImplementedError("Cannot call this method on base class.")

    @classmethod
    def run(cls, args: Namespace) -> None:
        """Run the command, generating any items and output them to stdout."""
        items: Generator[ItemT, None, None] = cls._generate(args)
        format_function = cls.output_function_by_format(args.format)
        sys.stdout.write(format_function(args, items) + "\n")


class DupeCommand(Command[Dupe]):
    """Run the dupe finding command."""

    @classmethod
    def json_output(cls, args: Namespace, items: Generator[Dupe, None, None]) -> str:
        """Outputs a json format string."""
        return json.dumps(
            {
                "algorithm": args.algorithm.algo_name,
                "dupes":
                    [
                        {
                            "image_hash": str(dupe.image_hash),
                            "paths": [str(path.resolve()) for path in dupe.paths]
                        }
                        for dupe in items
                    ]
            },
            indent=2,
        )

    @classmethod
    def text_output(cls, args: Namespace, items: Generator[Dupe, None, None]) -> str:
        """Outputs a text format string."""
        text_lines: List[str] = []
        for dupe in items:
            text_lines.append(str(dupe.image_hash))
            for path in dupe.paths:
                text_lines.append(str(path.resolve()))
        return "\n".join(text_lines)

    @classmethod
    def _generate(cls, args: Namespace) -> Generator[Dupe, None, None]:
        """Call the compare function with parsed args."""
        return Dupe.find(
            search_paths=args.search_paths,
            algorithm=args.algorithm,
        )


class CompareCommand(Command[ImageDiff]):
    """Run the compare command."""

    @classmethod
    def json_output(cls, args: Namespace, items: Generator[ImageDiff, None, None]) -> str:
        """Outputs a json format string."""
        return json.dumps(
            {
                "reference_path": str(args.ref_path.resolve()),
                "algorithm": args.algorithm.algo_name,
                "diffs":
                    [
                        {
                            "diff": image_diff.diff,
                            "path": str(image_diff.path.resolve())
                        }
                        for image_diff in items
                    ],
            },
            indent=2,
        )

    @classmethod
    def text_output(cls, args: Namespace, items: Generator[ImageDiff, None, None]) -> str:
        """Outputs a text format string."""
        text_lines: List[str] = [str(args.ref_path.resolve())]
        for image_diff in items:
            text_lines.append(
                f"{image_diff.diff}\t{image_diff.path.resolve()}")
        return "\n".join(text_lines)

    @classmethod
    def _generate(cls, args: Namespace) -> Generator[ImageDiff, None, None]:
        """Call the compare function with parsed args."""
        return ImageDiff.compare(
            ref_path=args.ref_path,
            search_paths=args.search_paths,
            algorithm=args.algorithm,
            threshold=args.threshold,
            stop_on_first_match=args.stop_on_first_match
        )



"""Functionality around searching paths and determining visual differences."""
from __future__ import annotations

from pathlib import Path
from typing import Generator, List, Optional

import attr

from .exceptions import ThresholdException
from .fingerprint import Algorithm, ImageFingerprint


@attr.s(frozen=True, auto_attribs=True, kw_only=True, order=True)
class ImageDiff:
    """The difference in hash between this image and a reference image."""
    path: Path = attr.ib(order=False)
    diff: int

    # def text_output(self) -> str:
    #     """Return the text string of this object"""
    #     return f"{self.diff}\t{self.path}"

    # def jsonable_output(self) -> Dict[str, Any]:
    #     """Return an object that can be json-encoded."""
    #     return dict(
    #         path=str(self.path),
    #         diff=self.diff
    #     )

    @staticmethod
    def _check_threshold(threshold: Optional[int]) -> None:
        """Checks the threshold for nonnegativity."""
        if threshold is not None and threshold < 0:
            raise ThresholdException("Threshold must be nonnegative.")


    @staticmethod
    def _check_stop_on_first_match(stop_on_first_match: bool, threshold: Optional[int]) -> None:
        """
        Checks that we can stop on the first match, which is only possible if a threshold is
        specified.
        """
        if threshold is None and stop_on_first_match:
            raise ThresholdException(
                "If stopping on first match, a threshold must be specified."
            )

    @classmethod
    def compare(
            cls,
            ref_path: Path,
            search_paths: List[Path],
            algorithm: Algorithm,
            stop_on_first_match: bool = False,
            threshold: Optional[int] = None,
    ) -> Generator[ImageDiff, None, None]:
        """
        Returns a generator of ImageDiff objects for all images in search paths against the ref path
        image.
        """
        cls._check_threshold(threshold)
        cls._check_stop_on_first_match(stop_on_first_match, threshold)

        reference_fingerprint = ImageFingerprint.from_path(ref_path, algorithm)

        for image_fingerprint in ImageFingerprint.recurse_paths(search_paths, algorithm):
            diff = image_fingerprint.image_hash - reference_fingerprint.image_hash
            if threshold is None or diff <= threshold:
                yield ImageDiff(
                    path=image_fingerprint.path,
                    diff=diff
                )
                if stop_on_first_match:
                    break



"""Compute visual fingerprints of images."""
from __future__ import annotations

from pathlib import Path
from typing import Callable, Dict, Generator, Iterable, Set
from enum import Enum, unique

import imagehash
from PIL import Image, UnidentifiedImageError
import attr

from .exceptions import (
    UnsupportedImageException,
    DoesNotExistException,
    NotReadableException,
    UnknownAlgorithmException,
    HashingException,
)


@unique
class Algorithm(Enum):
    """
    An enumeration of the algorithms provided by imagehash, along with a name and description.
    """
    AHASH = (
        'ahash',
        imagehash.average_hash,
        'Average hashing'
    )
    PHASH = (
        'phash',
        imagehash.phash,
        '2-axis perceptual hashing'
    )
    PHASH_SIMPLE = (
        'phash-simple',
        imagehash.phash_simple,
        '1-axis perceptual hashing'
    )
    DHASH = (
        'dhash',
        imagehash.dhash,
        'Horizontal difference hashing'
    )
    DHASH_VERT = (
        'dhash-vert',
        imagehash.dhash_vertical,
        'Vertical difference hashing'
    )
    WHASH_HAAR = (
        'whash-haar',
        imagehash.whash,
        'Haar wavelet hashing'
    )
    WHASH_DB4 = (
        'whash-db4',
        lambda img: imagehash.whash(img, mode='db4'),
        'Daubechies wavelet hashing'
    )
    COLORHASH = (
        'colorhash',
        imagehash.colorhash,
        'HSV color hashing'
    )

    @classmethod
    def from_name(cls, name: str) -> Algorithm:
        """Returns an algorithm by name."""
        name_map: Dict[str, Algorithm] = {
            algo.algo_name: algo for algo in list(Algorithm)
        }
        name_normalized = name.lower()
        if name_normalized not in name_map:
            raise UnknownAlgorithmException(f"No algorithm with name {name}")
        return name_map[name_normalized]

    @classmethod
    def supported_names(cls) -> str:
        """Returns a comma-separated string of all the algo names."""
        return ", ".join(algo.algo_name.lower() for algo in list(Algorithm))

    @classmethod
    def all_descriptions(cls) -> str:
        """Returns a comma-separated list of algorithm names and what they do."""
        return ", ".join(
            f"{algo.algo_name} = {algo.description}" for algo in list(Algorithm)
        )

    def __init__(
            self,
            algo_name: str,
            method: Callable[[Image], imagehash.ImageHash],
            description: str,
    ):
        self.algo_name = algo_name
        self.method = method
        self.description = description

    def hash_path(self, path: Path) -> imagehash.ImageHash:
        """
        Compute the visual fingerprint of image at path with an algorithm.

        Pillow accepts many, many image formats. Instead of trying to analyze what's acceptable
        early, we just attempt to open the path with it and handle exceptions appropriately.
        """
        try:
            image = Image.open(path)
        except UnidentifiedImageError as exc:
            raise UnsupportedImageException(
                f"Path {path} is not a supported image format: {exc}"
            )
        except FileNotFoundError as exc:
            raise DoesNotExistException(
                f"Path {path} could not be found: {exc}"
            )
        except PermissionError as exc:
            raise NotReadableException(
                f"Could not open path {path} for reading: {exc}"
            )
        except IsADirectoryError as exc:
            raise NotReadableException(
                f"Path {path} is a directory: {exc}"
            )
        return self.method(image)


@attr.s(frozen=True, auto_attribs=True, kw_only=True, order=False)
class ImageFingerprint:
    """
    A file that will be fingerprinted.

    path: The Path to the file
    explicit: Whether the file was explicitly asked to be fingerprinted (i.e. was provided as a
        search path and not a directory's child).
    """
    path: Path
    image_hash: imagehash.ImageHash
    algorithm: Algorithm

    @classmethod
    def from_path(cls, path: Path, algorithm: Algorithm) -> ImageFingerprint:
        """Create an ImageFingerprint from just a path (the hashing happens here)."""
        image_hash = algorithm.hash_path(path)

        return cls(
            path=path,
            image_hash=image_hash,
            algorithm=algorithm,
        )

    @classmethod
    def recurse_paths(
            cls,
            search_paths: Iterable[Path],
            algorithm: Algorithm,
    ) -> Generator[ImageFingerprint, None, None]:
        """
        Yields ImageFingerprint objects for all child paths in the iterable search_paths that are
        images.

        search_paths should be a list of Path objects. They may point to different types of path:
            - If the path is actually a directory, it will be recursed into and only hashable images
              yielded. This avoids any issues with hidden files, files that aren't images, etc.
            - If the path is a file, and it is not hashable, an exception will be thrown. Otherwise,
              it is yielded.

        Only unique paths will be yielded.
        """
        seen_paths: Set[Path] = set()

        for search_path in search_paths:

            if search_path.is_file():
                if search_path in seen_paths:
                    continue

                # this may throw a HashingException. we let it bubble up because the file was
                # explicitly named.
                yield ImageFingerprint.from_path(
                    path=search_path,
                    algorithm=algorithm,
                )

                seen_paths.add(search_path)

            elif search_path.is_dir():
                for child_path in search_path.rglob("*"):
                    if child_path.is_file() and child_path not in seen_paths:
                        try:
                            yield ImageFingerprint.from_path(
                                path=child_path,
                                algorithm=algorithm
                            )
                        except HashingException:
                            # here, we don't care about HashingExceptions because we're just
                            # scanning every child file. we don't require everything to be an image
                            # in a given dir. This is for usability (don't need to cleanse a dir to
                            # only contain images) and to workaround unavoidable hidden files
                            # (Thumbs.db, .DS_Store, etc)
                            continue
                        finally:
                            seen_paths.add(child_path)

            else:
                raise NotReadableException(
                    f'Search path "{search_path}" is not a file or directory that can be read.'
                )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imagesearch', 'imagesearch.cli']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0', 'imagehash>=4.1.0,<5.0.0']

entry_points = \
{'console_scripts': ['imagesearch = imagesearch.__main__:main']}

setup_kwargs = {
    'name': 'imagesearch',
    'version': '0.2.1',
    'description': 'Measure visual similiarity of a reference image to other images.',
    'long_description': '# imagesearch\n\n![PyPI](https://img.shields.io/pypi/v/imagesearch)\n![Build](https://github.com/t-mart/imagesearch/workflows/Build/badge.svg?branch=master)\n[![Coverage Status](https://coveralls.io/repos/github/t-mart/imagesearch/badge.svg?branch=master)](\nhttps://coveralls.io/github/t-mart/imagesearch?branch=master)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/imagesearch)\n\n`imagesearch` performs visual comparison operations on images.\n\nThe `compare` command measures visual similiarity between a reference image and a set of other\nimages. This can be used to search for a similar image that you know among a deep directory\nstructure of images that you don\'t want to manually scan.\n\nThe `dupe` command finds images which have identical visual fingerprints in a search path. This\ncan be used to identify images which you can delete later.\n\n## Installation\n\n    pip install imagesearch\n\nSee [imagesearch on PyPI](https://pypi.org/project/imagesearch/).\n\n## Examples\n\n- Get help:\n\n        > imagesearch --help\n        ...\n\n### Commands\n\n`imagesearch` functionality is broken up into subcommands provided on the command line.\n\nAll subcommands share a few arguments:\n\n- `-a`/`--algorithm` specifies which fingerprint algorithm to use. For the choice, see the section\n  below.\n- `-f`/`-format` specifies the output format of results. This can be either `text` or `json` (the\n  default). `json` should be used when the results are to be read by another program because\n  eccentricities in filenames will be properly encoded. _(All examples below use `text` for\n  clarity.)_\n\n### `search` Command\n\n**A `0` value indicates the highest level of similarity, or possibly a true match.**\n\n- Compare a reference image to all images in a search path:\n\n        > imagesearch search needle.jpg haystack\\ --format text\n        28      haystack\\0.jpg\n        38      haystack\\1.jpg\n        12      haystack\\2.jpg\n        18      haystack\\3.jpg\n        32      haystack\\4.jpg\n        29      haystack\\5.jpg\n        0       haystack\\6.jpg\n        29      haystack\\7.jpg\n        5       haystack\\8.jpg\n        28      haystack\\9.jpg\n\n    In this example, `haystack\\6.jpg` is most similar.\n\n- Compare against a single image:\n\n        > imagesearch search needle.jpg haystack\\1.jpg --format text\n        38       haystack\\1.jpg\n\n- Only return images with similarity less than or equal to 10:\n\n        > imagesearch search needle.jpg haystack\\ --threshold 10 --format text\n        0       haystack\\6.jpg\n        5       haystack\\8.jpg\n\n- Return the first image found under the threshold (0, in this case) and stop searching immediately:\n\n        > imagesearch search needle.jpg haystack\\ -t 0 --stop-on-first-match --format text\n        0       haystack\\6.jpg\n\n- Specify a different algorithm:\n\n        > imagesearch search needle.jpg haystack\\ --algorithm colorhash --format text\n        ...\n\n- Get more help:\n\n        > imagesearch search --help\n        ...\n\n### `dupe` Command\n\n- Find all visually similar images in a search path:\n\n        > imagesearch dupe images\\ --format text\n        fff7db9f03030203\n                images\\file-123.jpg\n                images\\deep\\subdir\\foo.jpeg\n        fcf8f0fae2c6c400\n                images\\a\\file-987.jpg\n                images\\subdir\\bar.png\n\n    Each set of paths that are similar is prefixed with its hash.\n\n- Get more help:\n\n        > imagesearch dupe --help\n        ...\n\n## Visual Similiarity\n\nAt its core, `imagesearch` creates image fingerprints and compares them to other fingerprints. A\ncritical feature of these fingerprints is that they can be numerically compared. **Images that are\ndifferent will have large differences in their fingerprints, and vice versa.**\n\nUnless you have a good understanding of the algorihms used, values should be treated as opaque and\nsubjective. It is dependent on the algorithm used to create the fingerprints and your criteria for\nwhat "similar" is.\n\nThis project uses the\n[imagehash](https://github.com/JohannesBuchner/imagehash) library to produce these fingerprints, and\nmore information about the techniques can be found there.\n\n## Algorithms\n\nAll the fingerprinting algorithms in `imagesearch` come from\n[imagehash](https://github.com/JohannesBuchner/imagehash). In `imagesearch`, you may specify which\nalgorithm to use by passing the appropriate option value to the `-a` or `--algorithm` flag:\n\n- `ahash`: Average hashing (aHash)\n- `phash`: 2-axis perceptual hashing (pHash)\n- `phash-simple`: 1-axis perceptual hashing (pHash)\n- `dhash`: Horizontal difference hashing (dHash)\n- `dhash-vert`: Vertical difference hashing (dHash)\n- `whash-haar`: Haar wavelet hashing (wHash)\n- `whash-db4`: Daubechies wavelet hashing (wHash)\n- `colorhash`: HSV color hashing (colorhash)\n\n## Collisions\nThese algorithms trade away accuracy for speed and size, usually with acceptable results. Instead of\nproducing an artifact that exactly identifies an image, there\'s analysis done on some more abstract\nquality of the image, such as it\'s luminance or\n[signal frequency](https://en.wikipedia.org/wiki/Discrete_cosine_transform). This allows us\nto:\n- do less processing\n- get a fingerprint with a small size\n- get a fingerprint that exists in a linear space for comparison\n\nHowever, because the exact image analysis is abstract and produces a fixed-size fingerprint, it\'s\nabsolutely possible for 2 different images to have the same fingerprint.\n\nThis is sort of an analog to cryptographic hash collosions, so it\'s important to understand what\nkinds of scenarios may cause this!\n\nSee\n[this section of the imagehash documentation](\nhttps://github.com/JohannesBuchner/imagehash#example-results) for examples of different images that\nproduce the same fingerprint. The source code of that project also references other pages that\nexplain the workings of the algorithm.\n\n## Contributing\n\n### Bug Fixes/Features\n\nSubmit a PR from an appropriately named feature branch off of master.\n\n### Releasing\n\n1. Bump the version with `bumpversion [patch|minor|major]`. This will update the version number\naround the project, commit and tag it.\n2. Push the repo. A Github release will be made and published to PyPI.\n',
    'author': 'Tim Martin',
    'author_email': 'tim@timmart.in',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/t-mart/imagesearch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
